from string import Template

"""Unused"""
ComponentTemplate = Template(
r"""<template>
</template>

<script setup lang="ts">
import {
  toRefs
} from 'vue'

import { useStore } from 'vuex'
import { useModule } from '@savitri/frontend'

const store = useStore()
const moduleRefs = reactive(useModule('${name}', store))

const {
  item,
  items

} = toRefs(moduleRefs)
</script>
""")

InternalHomeComponentTemplate = Template(
r"""<template>
  <sv-prose class="mt-4">
    <template #title>Parabéns</template>
    <ul class="flex flex-col gap-y-3 list-disc pl-4">
      <li
        v-for="(item, index) in items"
        v-html="item"
        :key="`item-$${index}`"
      />
    </ul>
  </sv-prose>
</template>

<script setup lang="ts">
import { inject } from 'vue'
import { SvProse } from '@savitri/components'

const productName = inject('productName')
const baseVersion = inject('baseVersion')

const items = [
  'Você configurou um projeto com sucesso',
  'Nome do projeto : ' + productName,
  'Versão do Savitri: ' + baseVersion
]
</script>
""")



from string import Template

ModelTemplate = Template(
r"""import { mongoose, options, descriptionToSchema } from '@savitri/backend'
import { default as Description } from './index.json'

${document_imports}
${reference_imports}

export interface ${pascal_case}Document extends mongoose.Document {
${type_hints}
}

export const ${pascal_case}Schema = descriptionToSchema<${pascal_case}Document>(Description, options)
export const ${pascal_case} = mongoose.model<${pascal_case}Document>('${pascal_case}', ${pascal_case}Schema)
""")




import re

MODULE_NAMING = [
    (lambda name: re.search(r'^[0-9]', name), 'module name cannot start with a number'),
    (lambda name: re.search(r'[A-Z]', name), 'module name must use snake_case'),
    (lambda name: name[-1:] == 's' and name[-2:] != 's', 'module name must be in singular')
]

FIELD_NAMING = [
    (lambda name: re.search(r'^[0-9]', name), 'field name cannot start with a number'),
    (lambda name: re.search(r'[A-Z]', name), 'field name must use snake_case'),
]

PLUGIN_NAMING = [
    (lambda name: not re.search(r'plugin-([a-z]+)$', name), 'module must follow "plugin-name" format')
]



from string import Template
from savitrigen.tree import TreeClass
from savitrigen.config import FrontendConfig
from savitrigen.template.frontend import (
    IndexTemplate,
    ModuleIndexTemplate,
    RouterTemplate,
    StoreTemplate,
    PluginImportTemplate,
    PluginInstanceTemplate,
    InternalRouterTemplate,
    InternalHomeComponentTemplate
)

@TreeClass('frontend')
class FrontendTree():
    def __init__(self, config:FrontendConfig):
        super().__init__()
        self._config = config

    def create(self):
        self.create_build_json()
        self.create_module('internal')

        self.write_file('index.ts', IndexTemplate, {
            'module_imports': self._multiline_replace(
                self._config.plugins,
                PluginImportTemplate,
                lambda _ : {
                    'capitalized': self._capitalize(_.split('-')[-1]),
                    'plugin_name': _
                }
            ),
            'module_instances': self._multiline_replace(
                self._config.plugins,
                PluginInstanceTemplate,
                lambda _ : {
                    'capitalized': ' '*4 + self._capitalize(_.split('-')[-1])
                }
            ),
            'menu_entries': self._multiline_replace(
                self._config.entities.keys(),
                Template(' '*8 + "'dashboard-${entity_name}',"),
                lambda _ : {
                    'entity_name': _
                }
            )
        })

        with self.change_dir('modules/internal'):
            self.write_file('router.ts', InternalRouterTemplate, {})
            self.make_dir('components/dashboard/c-home')
            self.write_file('components/dashboard/c-home/c-home.vue', InternalHomeComponentTemplate, {})

    def create_build_json(self):
        content = self._json_dumps({
            'name': '',
            'externals': {
                'variables': {
                    'productName': self._config.meta.product.name,
                    'productLogo': 'logo.png',
                    'productLogoAlt': 'logo-alt.png',
                    'releases': self._config.has_releases,
                    'notification': self._config.has_notification,
                    'feedback': self._config.has_feedback,

                }
            }
        })

        self.write_file('build.json', content)

    def create_module(self, name:str):
        path = self.make_dir('modules/{}'.format(name))

        store_path = self.make_dir('modules/{}/store'.format(name))
        components_path = self.make_dir('modules/{}/components'.format(name))

        with self.change_dir(path):
            self.write_file('index.ts', ModuleIndexTemplate, {})
            self.write_file('router.ts', RouterTemplate, {})
            self.write_file('store/index.ts', StoreTemplate, {})



from string import Template

RouterTemplate = Template(r"""
import { RouterExtension } from 'frontend/router'

export const routerExtension: RouterExtension = {
  //
}
""")

StoreTemplate = Template(r"""
import { StoreExtension } from 'frontend/store'

export const storeExtension: StoreExtension = {
  //
}
""")

InternalRouterTemplate = Template(r"""
import { RouterExtension } from '@savitri/frontend'

export const routerExtension: RouterExtension = {
  'dashboard': [
    {
      name: 'dashboard-home',
      path: 'home',
      component: () => import('./components/dashboard/c-home/c-home.vue'),
      meta: { title: 'Início' }
    }
  ]
}
""")



from string import Template

"""Used in models"""
DocumentImportTemplate = Template(
r"import { ${pascal_case}Document } from '../${camel_case}/${camel_case}.model'"
)

ReferenceImportTemplate = Template(
r"import '../${camel_case}/${camel_case}.model'"
)


"""Used in index"""
PluginImportTemplate = Template(
r"import { backend as ${capitalized}Module } from '$plugin_entity'"
)

PluginInstanceTemplate = Template(
r"${capitalized}Module(),"
)



from string import Template

PluginImportTemplate = Template(
r"""import { frontend as ${capitalized}Module } from '$plugin_name'"""
)

PluginInstanceTemplate = Template(
r"""${capitalized}Module(),"""
)



from .component import ComponentTemplate, InternalHomeComponentTemplate
from .index import ModuleIndexTemplate, IndexTemplate
from .extension import RouterTemplate, StoreTemplate, InternalRouterTemplate
from .imports import PluginImportTemplate, PluginInstanceTemplate



import pathlib
import contextlib
import logging
import typing
from shutil import copy
from string import Template
from multipledispatch import dispatch

class PathlibWrapper(object):
    def __init__(self):
        self.logger = logging.getLogger('path')
        self.parent_dir = None

    def _get_path(self, path) -> pathlib.Path:
        if isinstance(path, pathlib.Path):
            return path
        return pathlib.Path(path)

    @contextlib.contextmanager
    def change_dir(self, new_dir:str):
        previous_dir = self.parent_dir
        self.parent_dir = new_dir

        try:
            yield
        finally:
            self.parent_dir = previous_dir

    def make_dir(self, dirname:str) -> pathlib.Path:
        self.logger.info('MKDIR %s', dirname)
        path = self._get_path(dirname)
        path.mkdir(parents=True, exist_ok=True)
        return path

    def copy_file(self, src:str, dest:str) -> pathlib.Path:
        res = copy(src, '{}/{}'.format(self.parent_dir, dest))
        return self._get_path(res)

    def _write_file(self, fname:str, content:str) -> pathlib.Path:
        self.logger.info('WRITE %s', fname)
        path = self._get_path(fname)
        path.write_bytes(content.encode())
        return path

    @dispatch(str, str)
    def write_file(self, *args, **kwargs) -> pathlib.Path:
        return self._write_file(*args, **kwargs)

    @dispatch(str, Template, dict)
    def write_file(self, fname:str, template:Template, replacements:dict) -> pathlib.Path:
        return self._write_file(fname, template.substitute(**replacements))

    @dispatch(str, Template, dict, dict)
    def write_file(self, fname:str, template:Template, fallback:dict, replacements:dict) -> pathlib.Path:
        return self._write_file(fname, template.substitute(fallback, **replacements))



#!/usr/bin/env python3

import os
import runpy

from setuptools import setup, find_packages

def get_version():
    filename = os.path.join(os.path.dirname(__file__), 'savitrigen', '__init__.py')
    var = runpy.run_path(filename)
    return var['__version__']

_VERSION = get_version()

with open('README.rst', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='savitrigen',
    version=_VERSION,
    license='ISC',
    author='João Santos',
    author_email='joaosan177@gmail.com',
    description='Savitri code generator',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    scripts=['scripts/savitrigen'],
    include_package_data=True,
    package_data={
        'savitrigen': [
            'data/config.yml',
            'data/presets/*.yml'
        ]
    },
    install_requires=[
        'multipledispatch==0.6.0',
        'pyyaml==6.0'
    ]
)



from dataclasses import dataclass, field
from .backend import BackendConfig
from .frontend import FrontendConfig

from savitrigen.guideline import check_plugin_naming

@dataclass
class Product(object):
    name:str
    group:str

@dataclass
class Meta(object):
    version:str
    alias:str
    product:Product
    _product:Product = field(init=False, repr=False)

    @property
    def product(self) -> Product:
        return self._product

    @product.setter
    def product(self, value:dict) -> None:
        self._product = Product(**value)


@dataclass
class ProjectConfig(object):
    meta:dict
    _meta:dict = field(init=False, repr=False)
    plugins:list[str]
    _plugins:list[str] = field(init=False, repr=False)

    backend:BackendConfig = None
    frontend:FrontendConfig = None

    @property
    def meta(self) -> Meta:
        return self._meta

    @meta.setter
    def meta(self, value:dict) -> None:
        self._meta = Meta(**value)

    @property
    def plugins(self) -> list[str]:
        return self._plugins

    @plugins.setter
    def plugins(self, plugins:list[str]) -> None:
        if not isinstance(plugins, list):
            self._plugins = []
            return

        for p in plugins:
            check_plugin_naming(p)

        self._plugins = plugins



import yaml

import savitrigen.logging
from savitrigen.userdir import CONFIG_PATH, copy_from_userdir, get_presets
from savitrigen.tree.frontend import FrontendTree
from savitrigen.tree.backend import BackendTree
from savitrigen.tree.project import ProjectTree
from savitrigen.bootstrap import Bootstrap
from savitrigen.config import (
    CodegenConfig,
    ProjectConfig,
    BackendConfig,
    FrontendConfig
)

def main():
    """Main code generation routine"""
    with open('{}/{}'.format(CONFIG_PATH, 'config.yml'), 'r') as f:
        codegen_config = CodegenConfig(**yaml.safe_load(f))

    with open('briefing.yml', 'r') as f:
        project_config = ProjectConfig(**yaml.safe_load(f))
        backend_config = BackendConfig(
            **project_config.backend,
            plugins=project_config.plugins
        )

        frontend_config = FrontendConfig(
            **project_config.frontend,
            meta=project_config.meta,
            entities=backend_config.entities,
            plugins=project_config.plugins
        )

    Bootstrap.clone_repo(codegen_config)

    ProjectTree(project_config).create()
    BackendTree(backend_config).create()
    FrontendTree(frontend_config).create()

    Bootstrap.install()


def preset(name:str):
    """Copies ~/.savitrigen/presets/[name].yml to current directory"""
    copy_from_userdir('presets/{}.yml'.format(name), 'briefing.yml')

def list_presets():
    for p in get_presets():
        print(' - {}'.format(p))

def print_version():
    from savitrigen import __version__
    print(__version__)



from functools import reduce

ts_type_mapping = {
    'string': [
        'text',
        'string',
        'select',
        'checkbox',
        'readio'
    ],
    'number': [
        'number',
        'integer'
    ],
    'ObjectId': [
        'module',
        'objectid'
    ],
    'boolean': 'boolean',
    'Date': 'datetime',
    'any': 'object',
}

def extract_entity(field:dict) -> str:
    def extract_query(value:dict) -> dict:
        return value.get('__query', None) \
            if type(value) is dict else None

    is_array = field.get('array', False)

    if module := field.get('module'):
        return module, is_array

    if values := field.get('values'):
        if query := extract_query(
            values
            if not isinstance(field['values'], list)
            else field['values'][0]
        ):
            return (
                query['module'],
                is_array or isinstance(field['values'], list)
            )

def extract_entities(fields:dict) -> str:
    for k, v in fields.items():
        _tuple = extract_entity(v)
        if _tuple:
            yield _tuple

def convert_type(name:str) -> dict:
    return next((
        k
        for k, v in ts_type_mapping.items()
        if name in v
    ), None)

def map_fields(fields:dict) -> dict:
    def reducer(a:dict, item:tuple) -> dict:
        k, v = item
        field_type = v.get('type', 'text')

        _tuple = extract_entity(v)
        if _tuple:
            module, is_array = _tuple
            return a | { k: '{}Document{}'.format(
                module[0].upper() + module[1:],
                '[]' if is_array else ''
            ) }

        return a | { k: convert_type(field_type) }
    return reduce(reducer, fields.items(), {})

def make_ts_typehints(obj:dict) -> str:
    def reducer(a:str, item:tuple) -> list:
        k, v = item
        return a + ["  {}: {}".format(k, v)]
    return "\n".join(reduce(reducer, obj.items(), []))

